use std::sync::Arc;

use axum::{
    body::Body,
    extract::{Path, State},
    response::Redirect,
    routing::get,
    BoxError, Router,
};
use bytes::Bytes;
use http::{Request, Response};
use http_body_util::{combinators::UnsyncBoxBody, BodyExt};
use mime::Mime;
use sqlx::{query, PgPool};
use tower_http::services::ServeFile;
use tracing::{error, field, instrument};
use ulid::Ulid;
use uuid::Uuid;

use crate::{config::Config, error::AppError};

#[derive(Clone)]
struct SharedState {
    db: PgPool,
    config: Arc<Config>,
}

pub fn router(db: PgPool, config: Arc<Config>) -> Router {
    Router::new()
        .route("/:slug", get(redirect_link))
        .route("/f/:key", get(redirect_file))
        .with_state(SharedState { db, config })
}

async fn redirect_link(
    State(SharedState { db, .. }): State<SharedState>,
    Path(slug): Path<String>,
) -> Result<Redirect, AppError> {
    let result = query!("SELECT id, destination FROM link WHERE slug = $1", slug)
        .fetch_optional(&db)
        .await?
        .map(|r| (Ulid::from(r.id), r.destination));

    match result {
        Some((id, destination)) => {
            tokio::spawn(increase_visit_count(id, db));
            Ok(Redirect::temporary(&destination))
        }
        None => Err(AppError::LinkNotFound(slug)),
    }
}

#[instrument(skip(db))]
async fn increase_visit_count(id: Ulid, db: PgPool) {
    let result = query!(
        "UPDATE link SET visit_count = visit_count + 1 WHERE id = $1",
        Uuid::from(id),
    )
    .execute(&db)
    .await;

    match result {
        Ok(result) if result.rows_affected() != 1 => {
            error!(err = field::display(AppError::ImpossibleAffectedRows(result.rows_affected())));
        }
        Err(err) => error!(err = field::display(err)),
        _ => {}
    }
}

async fn redirect_file(
    State(SharedState { db, config }): State<SharedState>,
    Path(key): Path<Ulid>,
    request: Request<Body>,
) -> Result<Response<UnsyncBoxBody<Bytes, BoxError>>, AppError> {
    let result = query!(
        "SELECT file_hash, mime FROM file_key JOIN file ON file_hash = hash WHERE id = $1",
        Uuid::from(key)
    )
    .fetch_optional(&db)
    .await?
    .map(|r| (r.file_hash, r.mime));

    match result {
        Some((file_hash, mime)) => {
            let mime: Option<Mime> = mime.map_or(None, |m| m.parse().ok());
            let file_path = config.file_store_dir.join(hex::encode(file_hash));
            let mut sf = match mime {
                Some(mime) => ServeFile::new_with_mime(file_path, &mime),
                None => ServeFile::new(file_path),
            };
            match sf.try_call(request).await {
                Ok(response) => Ok(response.map(|body| body.map_err(Into::into).boxed_unsync())),
                Err(err) => Err(AppError::Io(err)),
            }
        }
        None => Err(AppError::FileKeyNotFound(key)),
    }
}
