use axum::{body::Body, response::IntoResponse};
use http::StatusCode;
use tracing::{error, field};
use ulid::Ulid;

#[derive(Debug, thiserror::Error)]
pub enum AppError {
    #[error("link already exists ({0})")]
    ApiLinkExists(Ulid),
    #[error("link not found ({0})")]
    ApiLinkNotFound(Ulid),
    #[error("link not found ({0})")]
    LinkNotFound(String),
    #[error("file key not found ({0})")]
    FileKeyNotFound(Ulid),
    #[error("database returned an impossible number of affected rows ({0})")]
    ImpossibleAffectedRows(u64),
    #[error("database error")]
    Database(#[from] sqlx::Error),
    #[error(transparent)]
    Io(#[from] std::io::Error),
    #[error(transparent)]
    Other(#[from] eyre::Report),
}

impl IntoResponse for AppError {
    fn into_response(self) -> axum::http::Response<Body> {
        error!(err = field::display(&self));
        match self {
            Self::ApiLinkExists(_) => (StatusCode::BAD_REQUEST, "Link already exists"),
            Self::ApiLinkNotFound(_) | Self::LinkNotFound(_) => {
                (StatusCode::NOT_FOUND, "Link not found")
            }
            Self::FileKeyNotFound(_) => (StatusCode::NOT_FOUND, "File key not found"),
            Self::ImpossibleAffectedRows(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "Database returned an impossible number of affected rows",
            ),
            Self::Database(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "A database error has occured",
            ),
            Self::Io(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "An I/O error has occured",
            ),
            Self::Other(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "An unknown error has occured",
            ),
        }
        .into_response()
    }
}
