mod app;
mod config;
mod error;

use eyre::Context;
use figment::{
    providers::{Env, Format, Toml},
    Figment,
};
use tokio::{net::TcpListener, runtime::Runtime};
use tracing::Level;
use tracing_subscriber::{filter, layer::SubscriberExt, util::SubscriberInitExt};

use self::app::build_app;
use crate::config::Config;

fn main() -> eyre::Result<()> {
    let filter = filter::Targets::new()
        .with_target("ncpn", Level::TRACE)
        .with_target("tower_http::trace::on_response", Level::TRACE)
        .with_target("tower_http::trace::on_request", Level::TRACE)
        .with_default(Level::INFO);

    let tracing_layer = tracing_subscriber::fmt::layer();

    tracing_subscriber::registry()
        .with(tracing_layer)
        .with(filter)
        .try_init()
        .map_err(eyre::Error::msg)
        .context("failed to initialize tracing subscriber")?;

    let config: Config = Figment::new()
        .merge(Toml::file("config.toml"))
        .merge(Env::raw())
        .extract()
        .context("failed to parse config")?;

    let rt = Runtime::new().context("failed to create tokio runtime")?;

    rt.block_on(async move {
        let listen_addr = config.listen_addr;

        let app = build_app(config).await.context("failed to build app")?;
        let listener = TcpListener::bind(&listen_addr)
            .await
            .context("failed to bind listener")?;

        axum::serve(listener, app)
            .await
            .context("server encountered a runtime error")?;

        Ok(())
    })
}
